import logging

import ckan.plugins as p
import ckan.plugins.toolkit as tk
from ckan import model

from ckanext.hierarchy.logic import action
from ckanext.hierarchy import helpers

log = logging.getLogger(__name__)


class HierarchyPlugin(p.SingletonPlugin):
    p.implements(p.IConfigurer, inherit=True)
    p.implements(p.IActions)
    p.implements(p.ITemplateHelpers)
    p.implements(p.IPackageController, inherit=True)

    # IConfigurer

    def update_config(self, config):
        tk.add_template_directory(config, "templates")
        tk.add_public_directory(config, "assets")
        tk.add_resource("assets", "hierarchy")

    # IActions

    def get_actions(self):
        return action.get_actions()

    # ITemplateHelpers
    def get_helpers(self):
        return helpers.get_helpers()

    # IPackageController

    def before_search(self, search_params):
        """When searching an organization, optionally extend the search any
        sub-organizations too. This is achieved by modifying the search options
        before they go to SOLR.
        """
        if not tk.h.hierarchy_enable_tree_search():
            return search_params

        if "ext_hierarchy_enable_tree_search" not in tk.request.params:
            return search_params

        fq = search_params.get("fq")

        group = model.Group.get(tk.request.view_args["id"])
        if not group:
            return search_params

        children_org_hierarchy = group.get_children_group_hierarchy(type="organization")
        children_names = [org[1] for org in children_org_hierarchy]

        if children_names:
            # remove existing owner_org:"<parent>" clause - we'll replace
            # it with the tree of orgs in a moment.
            owner_org_q = 'owner_org:"{}"'.format(group.id)
            expanded_org_q = "organization:({})".format(
                " OR ".join(
                    org_name
                    for org_name in [group.name]
                    + children_names
                )
            )

            search_params["fq"] = fq.replace(owner_org_q, expanded_org_q)

        return search_params



from bisect import bisect_right


class GroupTreeNode(dict):
    """Represents a group in a tree, used when rendering the tree.

    Is a dict, with links to child GroupTreeNodes, so that it is already
    'dictized' for output from the logic layer.
    """

    def __init__(self, group_dict):
        dict.__init__(self)
        self.update(group_dict)
        self["highlighted"] = False
        self["children"] = []
        # self._children_titles has a 1:1 relationship with values in
        # self.['children'], and is used to help keep them sorted by title
        self._children_titles = []

    def add_child_node(self, child_node):
        """Adds the child GroupTreeNode to this node, keeping the children
        in alphabetical order by title.
        """
        title = child_node["title"]
        insert_index = bisect_right(self._children_titles, title)
        self["children"].insert(insert_index, child_node)
        self._children_titles.insert(insert_index, title)

    def highlight(self):
        """Flag this group to indicate it should be shown highlighted
        when rendered."""
        self["highlighted"] = True


def group_dictize(group):
    """Convert a Group object into a dict suitable for GroupTreeNode
    Much simpler and quicker to do this than run than the full package_show.
    """
    return {
        "id": group.id,
        "name": group.name,
        "title": group.title,
        "type": group.type,
    }



import ckan.plugins.toolkit as tk
import ckan.model as model

CONFIG_TREE_SEARCH_EDNPOINTS = "ckanext.hierarchy.tree_search_endpoints"
DEFAULT_TREE_SEARCH_ENDPOINTS = "organization.read"


def get_helpers():
    return {
        "group_tree": group_tree,
        "group_tree_section": group_tree_section,
        "group_tree_parents": group_tree_parents,
        "group_tree_get_longname": group_tree_get_longname,
        "group_tree_highlight": group_tree_highlight,
        "get_allowable_parent_groups": get_allowable_parent_groups,
        "hierarchy_enable_tree_search": hierarchy_enable_tree_search,
    }


def group_tree(organizations=[], type_="organization"):
    full_tree_list = tk.get_action("group_tree")({}, {"type": type_})

    if not organizations:
        return full_tree_list
    else:
        filtered_tree_list = group_tree_filter(organizations, full_tree_list)
        return filtered_tree_list


def group_tree_filter(organizations, group_tree_list, highlight=False):
    # this method leaves only the sections of the tree corresponding to the
    # list since it was developed for the users, all children organizations
    # from the organizations in the list are included
    def traverse_select_highlighted(group_tree, selection=[], highlight=False):
        # add highlighted branches to the filtered tree
        if group_tree["highlighted"]:
            # add to the selection and remove highlighting if necessary
            if highlight:
                selection += [group_tree]
            else:
                selection += group_tree_highlight([], [group_tree])
        else:
            # check if there is any highlighted child tree
            for child in group_tree.get("children", []):
                traverse_select_highlighted(child, selection)

    filtered_tree = []
    # first highlights all the organizations from the list in the three
    for group in group_tree_highlight(organizations, group_tree_list):
        traverse_select_highlighted(group, filtered_tree, highlight)

    return filtered_tree


def group_tree_section(
    id_, type_="organization", include_parents=True, include_siblings=True
):
    return tk.get_action("group_tree_section")(
        {
            "include_parents": include_parents,
            "include_siblings": include_siblings,
        },
        {
            "id": id_,
            "type": type_,
        },
    )


def group_tree_parents(id_, type_="organization"):
    tree_node = tk.get_action("organization_show")({}, {"id": id_})
    if tree_node["groups"]:
        parent_id = tree_node["groups"][0]["name"]
        parent_node = tk.get_action("organization_show")({}, {"id": parent_id})
        return group_tree_parents(parent_id) + [parent_node]
    else:
        return []


def group_tree_get_longname(id_, default="", type_="organization"):
    tree_node = tk.get_action("organization_show")({}, {"id": id_})
    longname = tree_node.get("longname", default)
    if not longname:
        return default
    return longname


def group_tree_highlight(organizations, group_tree_list):
    def traverse_highlight(group_tree, name_list):
        group_tree["highlighted"] = group_tree.get("name", "") in name_list

        for child in group_tree.get("children", []):
            traverse_highlight(child, name_list)

    selected_names = [o.get("name", None) for o in organizations]

    for group in group_tree_list:
        traverse_highlight(group, selected_names)
    return group_tree_list


def get_allowable_parent_groups(group_id=None):
    if group_id:
        group = model.Group.get(group_id)
        allowable_parent_groups = group.groups_allowed_to_be_its_parent(
            type="organization"
        )
    else:
        allowable_parent_groups = model.Group.all(group_type="organization")
    return allowable_parent_groups


def hierarchy_enable_tree_search():
    endpoint = ".".join(tk.get_endpoint())
    supported_endpoints = tk.aslist(
        tk.config.get(
            CONFIG_TREE_SEARCH_EDNPOINTS, DEFAULT_TREE_SEARCH_ENDPOINTS
        )
    )
    return endpoint in supported_endpoints



import logging

import ckan.plugins.toolkit as tk
from ckanext.hierarchy.model import GroupTreeNode

log = logging.getLogger(__name__)


def get_actions():
    return {
        "group_tree": group_tree,
        "group_tree_section": group_tree_section,
    }


@tk.side_effect_free
def group_tree(context, data_dict):
    """Returns the full group tree hierarchy.

    :returns: list of top-level GroupTreeNodes
    """
    group_type = data_dict.get("type", "group")
    return [
        _group_tree_branch(group, type=group_type)
        for group in context["model"].Group.get_top_level_groups(type=group_type)
    ]


@tk.side_effect_free
def group_tree_section(context, data_dict):
    """Returns the section of the group tree hierarchy which includes the given
    group, from the top-level group downwards.

    :param id: the id or name of the group to include in the tree
    :param include_parents: if false, starts from given group
    :param include_siblingss: if false, excludes given group siblings
    :returns: the top GroupTreeNode of the tree section
    """
    group_name_or_id = tk.get_or_bust(data_dict, "id")
    group = context["model"].Group.get(group_name_or_id)
    if group is None:
        raise tk.ObjectNotFound
    group_type = data_dict.get("type", "group")
    if group.type != group_type:
        how_type_was_set = (
            "was specified"
            if data_dict.get("type")
            else "is filtered by default"
        )
        raise tk.ValidationError(
            'Group type is "%s" not "%s" that %s'
            % (group.type, group_type, how_type_was_set)
        )
    include_parents = context.get("include_parents", True)
    include_siblings = context.get("include_siblings", True)
    if include_parents:
        root_group = (
            group.get_parent_group_hierarchy(type=group_type) or [group]
        )[0]
    else:
        root_group = group
    if include_siblings or root_group == group:
        return _group_tree_branch(
            root_group, highlight_group_name=group.name, type=group_type
        )
    else:
        section_subtree = _group_tree_branch(
            group, highlight_group_name=group.name, type=group_type
        )
        return _nest_group_tree_list(
            group.get_parent_group_hierarchy(type=group_type), section_subtree
        )


def _nest_group_tree_list(group_tree_list, group_tree_leaf):
    """Returns a tree branch composed by nesting the groups in the list.

    :param group_tree_list: list of groups to build a tree, first is root
    :returns: the top GroupTreeNode of the tree
    """
    root_node = None
    last_node = None
    for group in group_tree_list:
        node = GroupTreeNode(
            {"id": group.id, "name": group.name, "title": group.title}
        )
        if not root_node:
            root_node = last_node = node
        else:
            last_node.add_child_node(node)
            last_node = node
    last_node.add_child_node(group_tree_leaf)
    return root_node


def _group_tree_branch(root_group, highlight_group_name=None, type="group"):
    """Returns a branch of the group tree hierarchy, rooted in the given group.

    :param root_group_id: group object at the top of the part of the tree
    :param highlight_group_name: group name that is to be flagged 'highlighted'
    :returns: the top GroupTreeNode of the tree
    """
    nodes = {}  # group_id: GroupTreeNode()
    root_node = nodes[root_group.id] = GroupTreeNode(
        {
            "id": root_group.id,
            "name": root_group.name,
            "title": root_group.title,
        }
    )
    if root_group.name == highlight_group_name:
        nodes[root_group.id].highlight()
        highlight_group_name = None
    for (
        group_id,
        group_name,
        group_title,
        parent_id,
    ) in root_group.get_children_group_hierarchy(type=type):
        node = GroupTreeNode(
            {"id": group_id, "name": group_name, "title": group_title}
        )
        nodes[parent_id].add_child_node(node)
        if highlight_group_name and group_name == highlight_group_name:
            node.highlight()
        nodes[group_id] = node
    return root_node
